from typing import Dict

from team_builder.nodes.interface import Iinputs
from .node import Node
from .enums import NodeType
import boto3
import json

class PythonNode(Node):
    code:str



    @property
    def type(self) -> NodeType:
        return NodeType.PYTHON

    def execute(self, inputs:Iinputs) -> str | Dict :
        return self.invoke_lambda(inputs.as_dict())

    @classmethod
    def from_kwargs(cls, **kwargs):
        return cls(
            id = kwargs["id"],
            name = kwargs["name"],
            code = kwargs["code"],
        )
    
    def invoke_lambda(self, inputs:Dict) -> Dict | str :
        lambda_client = boto3.client('lambda', region_name='eu-west-1')

        if self.code is None:
            return "Error: No script content provided."
            

        payload = {
            "script": self.code,
            "inputs": inputs,
        }

        response = lambda_client.invoke(
            FunctionName='executePythonScript',
            InvocationType='RequestResponse',
            Payload=json.dumps(payload)
        )

        res = json.load(response['Payload'])
        if res['status_code'] != 200:
            return res
        res['body'] = json.loads(res['body'])
        return res




from typing import Dict

from team_builder.nodes.enums import NodeType
from .node import Node
from .interface import Iinputs

class JumpNode(Node):
    jump_to_id:str
    jump_to_name:str

    @property
    def type(self) -> NodeType:
       return NodeType.JUMP
        
    def execute(self, inputs:Iinputs) -> str :
        return ""

    @classmethod
    def from_kwargs(cls, **kwargs):
        return cls(
            id = kwargs["id"],
            name = kwargs["name"],
            jump_to_id = kwargs["jump_to_id"],
            jump_to_name = kwargs["jump_to_name"],
        )



from enum import Enum
from typing import List, Any, Union, Dict, Iterator
from pydantic.v1 import BaseModel

from team_builder.nodes.enums import MessageType, NodeType, SchemaType
import re


class ISchema(BaseModel):
    key: str
    type: SchemaType
    description: str
    required: bool = True
    is_enum: bool = False
    enums: List[str] = []



class InputItem(BaseModel):
    key: str
    value: Any
    type: NodeType

class IHistory(BaseModel):
    text: str
    type: MessageType

class Iinputs(BaseModel):
    data: Dict[str, InputItem] = {}
    verbose: bool = True



    def __setitem__(self, key: str, item: InputItem) -> None:
        self.data[key] = item  # Use the data attribute directly

    def __getitem__(self, key: str) -> InputItem:
        return self.data[key]
    
    def get_value(self, key: str) -> str:
        """Returns the InputItem associated with the given key."""
        return str(self.data[key].value)

    def __iter__(self) -> Iterator[str]:
        return iter(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def as_dict(self) -> Dict[str, Any]:
        def serialize_value(value: Any) -> Any:
            if isinstance(value, BaseModel):
                return value.dict()  # Pydantic model, use its dict() method
            elif isinstance(value, list):
                return [serialize_value(item) for item in value]  # Recursively handle lists
            elif isinstance(value, Enum):  # Handle Enums by converting to their value
                return value.value
            else:
                return value  # Return the value as-is for other types

        return {k: serialize_value(v.value) for k, v in self.data.items()}
    
    def as_full_dict(self) -> Dict[str, Any]:
        return {k: {"value": v.value, "type":v.type} for k, v in self.data.items()}
    
    def as_dict_str(self) -> Dict[str, Union[str, Any]]:
        return {k: self.format_string(v) for k, v in self.data.items()}
    
    def format_string(self, inputs:InputItem) -> str | Dict:
        match inputs.type:
            case NodeType.HISTORY:
                return self.format_history(inputs.value)
            # case NodeType.RETRIVER:
            #     return self.get_doc_string(inputs.value)
            case _:
                return inputs.value
            
        
        



    def format_history(self, history:List[IHistory]) -> str:
        cleaned_history = []
        for item in history:
            text = item.text
            if self.verbose:
                text = re.sub(r'<(code|pre).*?>.*?</\1>', '', text, flags=re.DOTALL).strip()

            if MessageType.OUTBOUND:
                cleaned_history.append(f"AI: {text}")
            else:
                cleaned_history.append(f"Human: {text}")

        return "\n".join(cleaned_history)

       



















from typing import Dict

from team_builder.nodes.enums import NodeType
from .node import Node
from .interface import Iinputs

class TextNode(Node):
    text:str

    @property
    def type(self) -> NodeType:
       return NodeType.TEXT
        
    def execute(self, inputs:Iinputs) -> str :
        return self.convert_varibales(self.text, inputs)

    @classmethod
    def from_kwargs(cls, **kwargs):
        return cls(
            id = kwargs["id"],
            name = kwargs["name"],
            text = kwargs["text"],
        )



from typing import Optional

from OfficelyTestTeam2.generator import ThreadedGenerator
#from models.agents import agent
from .cohere import CohereLLM
from .google import GoogleLLM
from .tokenizer import Tokenizer
from .enums import LLMType
from .bedrock import BedrockLLM
from .openai import OpenAILLM
from .azure import AzureLLM
from .together import TogtherLLM
from OfficelyTestTeam2.CustomException import RootValidatorException




def factory_llm(
        llm_type:LLMType,
        temperature:float=0, 
        g: Optional[ThreadedGenerator]=None, 
        tokenizer:Optional[Tokenizer]=None,
        **kwargs
    ):
    """
    This is the factory_llm method.
    It creates a new instance of the LLM class based on the specified LLMType.
    """

    try:
        if is_openAI(llm_type):
            return OpenAILLM(model=llm_type, temperature=temperature, g=g, tokenizer=tokenizer)
        elif is_cohere(llm_type):
            return CohereLLM(model=llm_type, temperature=temperature, g=g, tokenizer=tokenizer)
        elif is_google(llm_type):
            return GoogleLLM(model=llm_type, temperature=temperature, g=g, tokenizer=tokenizer)
        elif llm_type == LLMType.AZURE:
            return AzureLLM(model=llm_type, temperature=temperature, g=g, tokenizer=tokenizer, **kwargs)
        elif llm_type == LLMType.TOGETHER:
            return TogtherLLM(model=llm_type, temperature=temperature, g=g, tokenizer=tokenizer, **kwargs)
        return BedrockLLM(model=llm_type, temperature=temperature, g=g, tokenizer=tokenizer, **kwargs)
    except RootValidatorException as e:
        raise e

def is_openAI(llm_type:LLMType):
    return llm_type.lower().startswith("gpt")

def is_anthropic(llm_type:LLMType):
    return llm_type.startswith("anthropic")

def is_google(llm_type:LLMType):
    return llm_type.startswith("gemini")


def is_cohere(llm_type:LLMType):
    return llm_type.startswith("command")





from OfficelyTestTeam2.stream_handler import StreamHandler
from langchain_google_genai import ChatGoogleGenerativeAI
from .llm import LLM
from langchain_core.language_models.chat_models import generate_from_stream
from typing import List, Optional, Sequence
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_google_genai._function_utils import _ToolConfigDict

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from urllib.parse import urlparse

from google.generativeai.types import Tool as GoogleTool  # type: ignore[import]
from google.generativeai.types.content_types import (  # type: ignore[import]
    FunctionDeclarationType,
    ToolDict,
)
from langchain_core.callbacks.manager import (CallbackManagerForLLMRun,)

from langchain_core.messages import (BaseMessage,)


from langchain_core.outputs import ChatResult

from langchain_google_genai._common import (SafetySettingDict,)
from langchain_google_genai._function_utils import ( _ToolConfigDict)




class GoogleGenAI(ChatGoogleGenerativeAI):
    """Create a custom class for GoogleGenAI that can handle streaming."""

    streaming:bool


    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        *,
        tools: Optional[Sequence[Union[ToolDict, GoogleTool]]] = None,
        functions: Optional[Sequence[FunctionDeclarationType]] = None,
        safety_settings: Optional[SafetySettingDict] = None,
        tool_config: Optional[Union[Dict, _ToolConfigDict]] = None,
        generation_config: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self.streaming:
            stream_iter = self._stream(
                    messages, stop=stop, run_manager=run_manager, **kwargs
                )
            return generate_from_stream(stream_iter)
        else:
            return super()._generate(messages, stop=stop, run_manager=run_manager, 
                              tools=tools, 
                              functions=functions,
                               safety_settings=safety_settings,
                               tool_config=tool_config,
                                generation_config=generation_config,
                                **kwargs
                            )


class GoogleLLM(LLM):

    @property
    def base(self):
        callbacks = []
        if self.tokenizer:
            self.tokenizer.model = self.model
            callbacks.append(self.tokenizer)
        is_answer = False
        if self.g:
            callbacks.append(StreamHandler(gen=self.g, with_final=False))
            is_answer = self.g.is_answer

        return GoogleGenAI( 
            model=self.model, 
            temperature=self.temperature, 
            callbacks=callbacks,
            streaming=is_answer,
            #type: ignore
        ) 

    
    
    def with_schema(self, prompt:List[Tuple[str, str]], schema:Dict, name:str):
        pass

    
   




from abc import ABC, abstractmethod
from typing import List, Optional
from pydantic.v1 import BaseModel, Field

from OfficelyTestTeam2.generator import ThreadedGenerator
from team_builder.nodes.interface import ISchema
from .enums import LLMType
from .tokenizer import Tokenizer
from langchain_core.prompt_values import PromptValue


class LLM(BaseModel, ABC):
    """
    This is the LLM class.
    It is a subclass of BaseModel and ABC.
    """
    model:LLMType
    temperature:float = 0
    g: Optional[ThreadedGenerator] = Field(default=None)
    tokenizer: Optional[Tokenizer] = Field(default=None) 




    @property
    @abstractmethod
    def base(self):
        """
        This is the base method.
        It serves as a placeholder and does not perform any specific functionality.
        """
    


    @abstractmethod
    def with_schema(self, prompt:str, schema:List[ISchema], name:str):
        """
        This is the with_structured_output method.
        It takes in a schema and returns a response.
        """









from team_builder.nodes.filter.filter_node import FilterNode
from team_builder.nodes.jump import JumpNode
from team_builder.nodes.python_node import PythonNode
from .enums import  EmbeddingModel, NodeType
from .llm_node import LLMNode
from .retriver_node import RetriverNode
from .text_node import TextNode



def factory_node(node_type:NodeType, **kwargs):
    """
    This is the factory_node method.
    It creates a new instance of the Node class based on the specified NodeType.
    """
    kwargs['name'] = kwargs['name'] if kwargs['name'] else kwargs['id']

    match node_type.lower():
        case NodeType.FILTER:
            return FilterNode.from_kwargs(**kwargs)
        case NodeType.LLM:
            return LLMNode.from_kwargs(**kwargs)
        case NodeType.RETRIVER:
            return RetriverNode.from_kwargs(**kwargs)
        case NodeType.PYTHON:
            return PythonNode.from_kwargs(**kwargs)
        case NodeType.TEXT:
            return TextNode.from_kwargs(**kwargs)
        case NodeType.JUMP:
            return JumpNode.from_kwargs(**kwargs)
        case _:
            raise ValueError(f"Invalid NodeType: {node_type}")
        



from enum import Enum

class LLMType(str, Enum):
    AZURE = "AZURE"
    TOGETHER = "TOGETHER",
    GPT3_5_TURBO = "gpt-3.5-turbo"
    GPT4 = "gpt-4"
    GPT_4_TURBO = "gpt-4-turbo"
    GPT4_O = "gpt-4o"
    GPT4_O_MINI = "gpt-4o-mini"
    CLAUDE3_SONNET = "anthropic.claude-3-sonnet-20240229-v1:0"
    CLAUDE3_HAIKU = "anthropic.claude-3-haiku-20240307-v1:0"
    CLAUDE3_5 = "anthropic.claude-3-5-sonnet-20240620-v1:0"
    CLAUDE2_1 = "anthropic.claude-v2:1"
    CLAUDE_INSTANT = "anthropic.claude-instant-v1"
    MISTRAL_7B = "mistral.mistral-7b-instruct-v0:2"
    MISTRAL_8X7B = "mistral.mixtral-8x7b-instruct-v0:1"
    MISTRAL_LARGE = "mistral.mistral-large-2402-v1:0"
    MISTRAL_SMALL = "mistral.mistral-small-2402-v1:0"
    LLAMA3_8B = "meta.llama3-8b-instruct-v1:0"
    LLAMA3_70B = "meta.llama3-70b-instruct-v1:0"
    LLAMA2_13B = "meta.llama2-13b-chat-v1"
    LLAMA2_70B = "meta.llama2-70b-chat-v1"
    # COMMAND_R_PLUS = "cohere.command-r-plus-v1:0"
    # COMMAND_R = "cohere.command-r-v1:0"
    # COMMAND = "cohere.command-text-v14"
    # COMMAND_LIGHT = "cohere.command-text-v14"
    COMMAND_R_PLUS = "command-r-plus"
    COMMAND_R = "command-r"
    COMMAND = "command"
    COMMAND_LIGHT = "command-light"
    # JURASSIC2_ULTRA = "ai21.j2-ultra-v1"
    # JURASSIC2_MID = "ai21.j2-mid-v1"
    GEMINI_1_5_PRO = "gemini-1.5-pro"
    
    



from langchain_cohere import ChatCohere
from pydantic.v1 import BaseModel, Field
from typing import Any, Dict , Generator, List, Optional, Tuple, Union
from tiktoken import encoding_for_model
from contextlib import contextmanager
from contextvars import ContextVar
from langchain_core.tracers.context import register_configure_hook
from langchain_community.utilities.anthropic import get_num_tokens_anthropic
from langchain.schema import LLMResult
from langchain.callbacks.base import BaseCallbackHandler
from langchain_community.callbacks.openai_info import get_openai_token_cost_for_model
from langchain_core.language_models.base import _get_token_ids_default_method
from team_builder.llms.enums import LLMType
from langchain_google_genai import ChatGoogleGenerativeAI



MODEL_COST_PER_1K_INPUT_TOKENS = {
    LLMType.MISTRAL_7B:0.00015,
    LLMType.MISTRAL_8X7B:0.00045,
    LLMType.MISTRAL_LARGE:0.008,
    LLMType.MISTRAL_SMALL:0.001,
    LLMType.LLAMA3_8B:0.0004,
    LLMType.LLAMA3_70B:0.00265,
    LLMType.COMMAND_R_PLUS:0.0030,
    LLMType.COMMAND_R:0.0005,
    LLMType.COMMAND:0.0015,
    LLMType.COMMAND_LIGHT:0.0003,
    LLMType.GEMINI_1_5_PRO: 0.0035
}

MODEL_COST_PER_1K_OUTPUT_TOKENS = {
    LLMType.MISTRAL_7B:0.0002,
    LLMType.MISTRAL_8X7B:0.0007,
    LLMType.MISTRAL_LARGE:0.024,
    LLMType.MISTRAL_SMALL:0.003,
    LLMType.LLAMA3_8B:0.0006,
    LLMType.LLAMA3_70B:0.0035,
    LLMType.COMMAND_R_PLUS:0.0150,
    LLMType.COMMAND_R:0.0015,
    LLMType.COMMAND:0.0020,
    LLMType.COMMAND_LIGHT:0.0006,
    LLMType.GEMINI_1_5_PRO: 0.0105
}


ANTH_MODEL_COST_PER_1K_INPUT_TOKENS = {
    "anthropic.claude-instant-v1": 0.0008,
    "anthropic.claude-v2": 0.008,
    "anthropic.claude-v2:1": 0.008,
    "anthropic.claude-3-sonnet-20240229-v1:0": 0.003,
    "anthropic.claude-3-haiku-20240307-v1:0": 0.00025,
    "anthropic.claude-3-5-sonnet-20240620-v1:0": 0.003,
}

ANTH_MODEL_COST_PER_1K_OUTPUT_TOKENS = {
    "anthropic.claude-instant-v1": 0.0024,
    "anthropic.claude-v2": 0.024,
    "anthropic.claude-v2:1": 0.024,
    "anthropic.claude-3-sonnet-20240229-v1:0": 0.015,
    "anthropic.claude-3-haiku-20240307-v1:0": 0.00125,
    "anthropic.claude-3-5-sonnet-20240620-v1:0": 0.015,
}


def _get_anthropic_claude_token_cost(
    prompt_tokens: int, completion_tokens: int, model_id: Union[str, None]
) -> float:
    """Get the cost of tokens for the Claude model."""
    if model_id not in ANTH_MODEL_COST_PER_1K_INPUT_TOKENS:
        raise ValueError(
            f"Unknown model: {model_id}. Please provide a valid Anthropic model name."
            "Known models are: " + ", ".join(ANTH_MODEL_COST_PER_1K_INPUT_TOKENS.keys())
        )
    return (prompt_tokens / 1000) * ANTH_MODEL_COST_PER_1K_INPUT_TOKENS[model_id] + (
        completion_tokens / 1000
    ) * ANTH_MODEL_COST_PER_1K_OUTPUT_TOKENS[model_id]


def _token_cost(
    prompt_tokens: int, completion_tokens: int, model_id: LLMType
) -> float:
    """Get the cost of tokens for the Claude model."""
    if model_id not in MODEL_COST_PER_1K_INPUT_TOKENS:
        # raise ValueError(
        #     f"Unknown model: {model_id}. Please provide a valid model name."
        #     "Known models are: " + ", ".join(MODEL_COST_PER_1K_INPUT_TOKENS.keys())
        # )
        return 0
    return (prompt_tokens / 1000) * MODEL_COST_PER_1K_INPUT_TOKENS[model_id] + (
        completion_tokens / 1000
    ) * MODEL_COST_PER_1K_OUTPUT_TOKENS[model_id]





class ITokenizer(BaseModel):
    successful_requests:int = 0
    total_tokens_input:int = 0
    total_tokens_output:int = 0
    total_cost:float = 0.0  

class ITokensData(BaseModel):
    model:LLMType
    outputs:int
    inputs:int


class Tokenizer(BaseModel,BaseCallbackHandler):
    model:Optional[LLMType] = None
    data:Dict[LLMType, ITokenizer] = {}
    last_token_input:int = Field(default=0, init_var=False)



    @property
    def Tokens_data(self) -> List[ITokensData]:
        tokens_data = []
        for model_key, tokenizer in self.data.items():
            prompts_tokens = tokenizer.total_tokens_input
            completion_tokens = tokenizer.total_tokens_output
            tokens_data.append(ITokensData(model=model_key, outputs=completion_tokens, inputs=prompts_tokens))
        return tokens_data
    
    def __str__(self) -> str:
        result = []
        for model_key, tokenizer in self.data.items():
            if model_key == LLMType.AZURE:
                continue
            prompts_tokens = tokenizer.total_tokens_input
            completion_tokens = tokenizer.total_tokens_output
            total_cost = self.calculate_cost(model_key, prompts_tokens, completion_tokens) #type: ignore
            self.data[model_key].total_cost = total_cost
            result.append(
                f"Model: {model_key.name if isinstance(model_key, LLMType) else model_key}\n"
                f"  Successful requests: {tokenizer.successful_requests}\n"
                f"  Total tokens input: {prompts_tokens}\n"
                f"  Total tokens output: {completion_tokens}\n"
                f"  Total cost: ${total_cost}\n"
            )
        result.append(f"Officely AI Fee: $0.01")
        result.append(f"Total Cost: ${0.01 + sum([tokenizer.total_cost for tokenizer in self.data.values()])}")
        return "\n".join(result)
    

    def start(self, content:str):
        if self.model is None:
            return
        if self.model not in self.data.keys():
            self.data[self.model] = ITokenizer()
        self.last_token_input = self.get_tokens_received(content=content)
        self.data[self.model].total_tokens_input += self.last_token_input

    def on_llm_start(self, serialized: Dict[str, Any],prompts: List[str], **kwarg):
        self.start(" ".join(prompts))
    
    def on_chat_model_start(self, serialized, messages, **kwargs):
        """Run when a chat model starts running."""
        def content(content:list | str):
            if isinstance(content, str):
                return content
            string = ""
            for x in content:
                if isinstance(x, str):
                    string += x
                if isinstance(x, dict) and x['type'] == "text": 
                    string += x['text']
            return string

        content_str = " ".join([content(x.content) for x in messages[0]])
        self.start(content_str)

    
    def get_tokens_received(self, content:Optional[str]=None, response:Any=None):
        """Get the number of tokens received in a message."""
        if response:
            message = response.generations[0][0].text
        else: 
            message = content
        num = 0
        if self.model.startswith("gpt"): #type: ignore
            if response and response.llm_output and "token_usage" in response.llm_output:
                token_usage = response.llm_output["token_usage"]
                num, _ = self.token_usage(token_usage)
            else:
                tiktoken = encoding_for_model(self.model) #type: ignore
                num = len(tiktoken.encode(message)) #type: ignore
        elif self.model.startswith("anthropic"): #type: ignore
            if response and response.llm_output and "usage" in response.llm_output:
                token_usage = response.llm_output["usage"]
                num, _ = self.token_usage(token_usage)
            else:
                num = get_num_tokens_anthropic(message)#type: ignore
        elif self.model.startswith("command"): #type: ignore
            num = ChatCohere(model=self.model).get_num_tokens(message) #type: ignore
        elif self.model.startswith("gemini"):    #type: ignore
            num = ChatGoogleGenerativeAI(model=self.model).get_num_tokens(message) #type: ignore
        else:
            num = len(_get_token_ids_default_method(message)) #type: ignore
        return num    
    
    def calculate_cost(self, model:LLMType, prompt_token:int, completion_tokens:bool):
        """Calculate the cost of tokens for a model."""
        cost = 0
        if model.startswith("gpt"):
            cost = get_openai_token_cost_for_model(model.value, prompt_token, False) + get_openai_token_cost_for_model(model.value, completion_tokens, True)
        elif model.startswith("anthropic"):
            cost =  _get_anthropic_claude_token_cost(prompt_token, completion_tokens, model.value)
        else:
            cost = _token_cost(prompt_token, completion_tokens, model)  
        return cost

    
    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        """Run when LLM ends running."""
        self.data[self.model].successful_requests += 1 #type: ignore
        self.data[self.model].total_tokens_output += self.get_tokens_received(response=response) #type: ignore


    def token_usage(self, token_usage:Dict[str, int]) -> Tuple[int, int]:
        completion_tokens = token_usage.get("completion_tokens", 0)
        prompt_tokens = token_usage.get("prompt_tokens", 0)
        if prompt_tokens:
            self.data[self.model].total_tokens_input -= self.last_token_input #type: ignore
            self.data[self.model].total_tokens_input += prompt_tokens #type: ignore
        return completion_tokens, prompt_tokens

    

tokenizer_callback_var: ContextVar[Optional[Tokenizer]] = ContextVar(
    "get_tokenizer_callback", default=None
)

register_configure_hook(tokenizer_callback_var, True)

    
@contextmanager
def get_team_tokenizer_callback(tokenizer: Tokenizer) -> Generator[Tokenizer, None, None]:
    """Context manager for managing a tokenizer callback.
    
    This is used to expose token and cost information from a tokenizer,
    typically used in NLP models.

    Args:
        tokenizer (Tokenizer): The tokenizer to be managed.

    Yields:
        Tokenizer: The managed tokenizer.
    """
    tokenizer_callback_var.set(tokenizer)
    try:
        yield tokenizer
    finally:
        tokenizer_callback_var.set(None)


from typing import Any, List
from dataclasses import dataclass, field
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler

from OfficelyTestTeam2.generator import ThreadedGenerator

DEFAULT_ANSWER_PREFIX_TOKENS = ["Final", "Answer", ":"]


@dataclass
class StreamHandler(StreamingStdOutCallbackHandler):
    gen: ThreadedGenerator
    with_final: bool
    answer_reached: bool = field(default=False, init=False)
    answer_prefix_tokens: List[str] = field(default_factory=lambda: DEFAULT_ANSWER_PREFIX_TOKENS)
    last_tokens: List[str] = field(default_factory=list, init=False)
    last_tokens_stripped: List[str] = field(default_factory=list, init=False)


    def __post_init__(self):
        super().__init__()


    def on_new_token(self, token: str) -> None:
        if self.with_final:
            self.final_stream(token)
        else:
            self.gen.send(token)
            
    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        if self.gen.team:
            if self.gen.is_answer:
                self.on_new_token(token)
        else:
            self.on_new_token(token)



    def append_to_last_tokens(self, token: str) -> None:
        self.last_tokens.append(token)
        self.last_tokens_stripped.append(token.strip())
        if len(self.last_tokens) > len(self.answer_prefix_tokens):
            self.last_tokens.pop(0)
            self.last_tokens_stripped.pop(0)

    def check_if_answer_reached(self) -> bool:
        return self.last_tokens_stripped == self.answer_prefix_tokens

    def final_stream(self, token:str):
        """Stream the final answer."""
        self.append_to_last_tokens(token)
        
        if not self.answer_reached:
            self.answer_reached = self.check_if_answer_reached()
            if self.answer_reached:
                return
            

        if self.answer_reached:
            self.gen.send(token)

    def on_chat_model_start(
        self,
        serialized,
        messages,
        **kwargs,
    ) -> None:
        """Run when LLM starts running."""
        self.answer_reached = False



    # def on_chat_model_start(
    #     self,
    #     serialized,
    #     messages,
    #     **kwargs,
    # ) -> None:
    #     """Run when LLM starts running."""
    #     self.model = kwargs["invocation_params"]["model"]
    #     self.gen.total_tokens_input += self.get_tokens_received(messages[0][0].content)


from dataclasses import dataclass
import queue



@dataclass
class ThreadedGenerator:
    team:bool = False
    is_answer:bool = False
    full_answer:str = ""

    def __post_init__ (self):
        self.queue = queue.Queue()
        


    def __iter__(self):
        return self

    def __next__(self):
        item = self.queue.get()
        if item is StopIteration:
            raise item
        return item

    def send(self, data, is_answer=True):
        if is_answer:
            self.full_answer += data
        self.queue.put(data)

    def close(self):
        self.queue.put(StopIteration)


from typing import Dict, List, Tuple

from langchain_together import ChatTogether

from OfficelyTestTeam2.stream_handler import StreamHandler
import requests
from .llm import LLM
import os




class TogtherLLM(LLM):
    togther_model:str


    streaming:bool = True

    @property
    def base(self):


        callbacks = []
        if self.tokenizer:
            self.tokenizer.model = self.togther_model #type:ignore
            callbacks.append(self.tokenizer)
        is_answer = False
        if self.g:
            callbacks.append(StreamHandler(gen=self.g, with_final=False))
            is_answer = self.g.is_answer

        return ChatTogether(
            model=self.togther_model, 
            temperature=self.temperature, 
            streaming=is_answer and self.streaming,
            callbacks=callbacks,
        )

    
    
    def with_schema(self, prompt:List[Tuple[str, str]], schema:Dict, name:str):
        pass
        # self.streaming = False
        # return self.base.with_structured_output(schema).with_config({"run_name":name}).invoke(prompt)

    @staticmethod
    def get_models_list():
        try:
            url = "https://api.together.xyz/v1/models"

            headers = {
                "accept": "application/json",
                "Authorization": f"Bearer {os.environ['TOGETHER_API_KEY']}"
            }

            response = requests.get(url, headers=headers).json()

            _list = [{"label":x['display_name'], "value":x['id']} for x in response if x['type'] == 'chat']
            sorted_list = sorted(_list, key=lambda x: x['label'])
            return sorted_list
        except Exception as e:
            raise e




    




from typing import Dict, List, Tuple
import boto3
from langchain_aws import ChatBedrock, BedrockLLM as Bedrock 

from OfficelyTestTeam2.stream_handler import StreamHandler
from team_builder.llms.enums import LLMType

from .llm import LLM





class BedrockLLM(LLM):
    region:str = "us-east-1"



    @property
    def Client(self):
        kwargs = {}


        kwargs.update({
            "service_name":'bedrock-runtime',
            "region_name": self.region
        })
        return boto3.client(**kwargs)

    
    @property
    def bedrock_chat_kwargs(self):
        callbacks = []
        if self.tokenizer:
            self.tokenizer.model = self.model
            callbacks.append(self.tokenizer)
        is_answer = False
        if self.g:
            callbacks.append(StreamHandler(gen=self.g, with_final=False))
            is_answer = self.g.is_answer
        return {
            "model_id": self.model,
            "model_kwargs": {"temperature": self.temperature},
            "client": self.Client,
            "callbacks": callbacks,
            "streaming": is_answer,

        }

    @property
    def base(self):
        if self.model in [LLMType.COMMAND_R_PLUS, LLMType.COMMAND_R, LLMType.COMMAND, LLMType.COMMAND_LIGHT]:
            return Bedrock(**self.bedrock_chat_kwargs)
        return ChatBedrock(**self.bedrock_chat_kwargs)

    
    
    def with_schema(self, prompt:List[Tuple[str, str]], schema:Dict, name:str):
        pass




    




from typing import Dict, List, Tuple


from OfficelyTestTeam2.stream_handler import StreamHandler
from langchain_cohere import ChatCohere

from .llm import LLM



from typing import Dict, List, Tuple




class CohereLLM(LLM):
    streaming:bool = True

    @property
    def base(self):
        callbacks = []
        if self.tokenizer:
            self.tokenizer.model = self.model
            callbacks.append(self.tokenizer)
        is_answer = False
        if self.g:
            callbacks.append(StreamHandler(gen=self.g, with_final=False))
            is_answer = self.g.is_answer

        return ChatCohere(
            model=self.model, 
            temperature=self.temperature, 
            streaming=is_answer and self.streaming,
            callbacks=callbacks,

        )

    
    
    def with_schema(self, prompt:List[Tuple[str, str]], schema:Dict, name:str):
        self.streaming = False
        try:
            cohere_schema = {
                'title': schema['name'],
                'description': schema['description'],
                'properties': schema['parameters']['properties'],
            }
            return self.base.with_structured_output(cohere_schema).with_config({"run_name":name}).invoke(prompt)
        except Exception as e:
            print(e)
            raise e
        

    
   


from .factory import factory_node
from .enums import NodeType
from .node import Node
from .llm_node import LLMNode
from .retriver_node import RetriverNode

__all__ = ["factory_node", "NodeType", "Node", "LLMNode", "RetriverNode"]


from enum import Enum

class DataType(str, Enum):
    TEXT = 'TEXT'
    NUMBER = 'NUMBER'
    BOOLEAN = 'BOOLEAN'
    DATE = 'DATE'
    BASIC = 'BASIC'


class OperatorLabel(str, Enum):
    EQUAL = 'equal to'
    NOT_EQUAL = 'not equal to '
    EQUAL_INSENSETIVE = 'equal to (case insensitive)'
    NOT_EQUAL_INSENSETIVE = 'not equal to (case insensitive)'
    CONTAINS_INSENSETIVE = 'contains (case insensitive)'
    NOT_CONTAINS_INSENSETIVE = 'does not contain (case insensitive)'
    CONTAINS = 'contains'
    NOT_CONTAINS = 'does not contain'
    STARTS_WITH = 'starts with'
    NOT_STARTS_WITH = 'does not start with'
    STARTS_WITH_INSENSETIVE = 'starts with (case insensitive)'
    NOT_STARTS_WITH_INSENSETIVE = 'does not start with (case insensitive)'
    ENDS_WITH_INSENSETIVE = 'ends with (case insensitive)'
    NOT_ENDS_WITH_INSENSETIVE= 'does not end with (case insensitive)'
    ENDS_WITH = 'ends with'
    NOT_ENDS_WITH = 'does not end with'
    GREATER_THAN = 'greater than'
    LESS_THAN = 'less than'
    GREATER_THAN_OR_EQUAL_TO = 'greater than or equal to'
    LESS_THAN_OR_EQUAL_TO = 'less than or equal to'
    MATCHES_REGEX = 'matches regex'
    NOT_MATCHES_REGEX = 'does not match regex'
    MATCHES_REGEX_INSENSETIVE = 'matches regex (case insensitive)'
    NOT_MATCHES_REGEX_INSENSETIVE = 'does not match regex (case insensitive)'
    EXISTS = 'exists'
    NOT_EXISTS = 'does not exist'


class OperatorValue(str, Enum):
    EQUAL = "equal to"
    NOT_EQUAL = "not equal to "
    EQUAL_INSENSETIVE = "equal to (case insensitive)"
    NOT_EQUAL_INSENSETIVE = "not equal to (case insensitive)"
    CONTAINS_INSENSETIVE = "contains (case insensitive)"
    NOT_CONTAINS_INSENSETIVE = "does not contain (case insensitive)"
    CONTAINS = "contains"
    NOT_CONTAINS = "does not contain"
    STARTS_WITH = "starts with"
    NOT_STARTS_WITH = "does not start with"
    STARTS_WITH_INSENSETIVE = "starts with (case insensitive)"
    NOT_STARTS_WITH_INSENSETIVE = "does not start with (case insensitive)"
    ENDS_WITH_INSENSETIVE = "ends with (case insensitive)"
    NOT_ENDS_WITH_INSENSETIVE = "does not end with (case insensitive)"
    ENDS_WITH = "ends with"
    NOT_ENDS_WITH = "does not end with"
    GREATER_THAN = "greater than"
    LESS_THAN = "less than"
    GREATER_THAN_OR_EQUAL_TO = "greater than or equal to"
    LESS_THAN_OR_EQUAL_TO = "less than or equal to"
    MATCHES_REGEX = "matches regex"
    NOT_MATCHES_REGEX = "does not match regex"
    MATCHES_REGEX_INSENSETIVE = "matches regex (case insensitive)"
    NOT_MATCHES_REGEX_INSENSETIVE = "does not match regex (case insensitive)"
    EXISTS = "exists"
    NOT_EXISTS = "does not exist"




import base64
from typing import Any, Dict, List, Optional, Tuple
import base64
import httpx
from langchain_core.output_parsers import StrOutputParser

from OfficelyTestTeam2.generator import ThreadedGenerator
from team_builder.llms import LLMType, factory_llm, is_openAI, is_anthropic
from team_builder.llms.tokenizer import Tokenizer
from .enums import NodeType
from .interface import ISchema, Iinputs
from .node import Node





class LLMNode(Node):
    """this node use langchain """
    model:LLMType
    system_prompt:str
    prompt:str
    with_schema:bool
    map_model:str
    llm_map:bool
    temperature:Optional[float] = None
    schema_data:Optional[List[ISchema]] = None
    g:Optional[ThreadedGenerator] = None
    tokenizer:Optional[Tokenizer] = None
    extra_kwargs:Dict[str, Any]



    @property
    def type(self):
        return NodeType.LLM
    
    @property
    def Allowed_schema(self):
        return self.model in [LLMType.GPT4, LLMType.GPT_4_TURBO, LLMType.GPT4_O, LLMType.GPT4_O_MINI, LLMType.AZURE,
            LLMType.COMMAND_R_PLUS, LLMType.COMMAND_R, LLMType.COMMAND, LLMType.COMMAND_LIGHT
        ]

    
    @classmethod
    def from_kwargs(cls, **kwargs):
        return cls(
            id = kwargs["id"],
            name = kwargs["name"],
            model = LLMType[kwargs["model"]],
            system_prompt = kwargs.get("system_prompt", ""),
            prompt = kwargs["prompt"],
            with_schema = kwargs.get("with_schema", False),
            description = kwargs.get("description"),
            temperature = kwargs.get("temperature", 0),
            schema_data = kwargs.get("schema"),
            map_model = str(kwargs.get("model_string", '')).strip(),
            llm_map = kwargs.get("llm_map", False),
            extra_kwargs = cls.get_extra_kwargs(**kwargs)
        )
    
    @classmethod
    def get_extra_kwargs(cls, **kwargs):
        return {
            "azure_deployment_name":kwargs.get("azureDeploymentName", None),
            "own_bedrock":kwargs.get("ownBedrock", False),
            "togther_model":kwargs.get("togtherModel", ""),
        }
    
    def is_valid_model(self, model: str) -> bool:
        return model in LLMType.__members__.values()

    def get_llm(self, inputs:Iinputs):
        if self.llm_map:
            model = self.convert_varibales(self.map_model, inputs)
            if self.is_valid_model(model):
                model = LLMType(model)
            else:
                err = f"model <b>{model}</b> is not valid you need to choose one of: <br><br> {'<br>'.join(LLMType.__members__.values())}"
                err += f"<br><br> from node <b>{self.name}<b>"
                raise ValueError(err)
        else:
            model = self.model
        return factory_llm(model, g=self.g, tokenizer=self.tokenizer, **self.extra_kwargs)


    def execute(self, inputs:Iinputs):
        if self.with_schema and self.Allowed_schema:
            return self.run_schema(inputs)
        return self.run(inputs)
    
    def run(self, inputs:Iinputs) -> str:
        llm = self.get_llm(inputs)
        PROMPT = self.create_messages(inputs)
        chain = llm.base | StrOutputParser()
        return chain.with_config({"run_name":self.name}).invoke(PROMPT) 
    
    def create_messages(self, inputs:Iinputs) -> List[Tuple[str, str]]:
        system = self.convert_varibales(self.system_prompt, inputs)
        messages = []
        if system:
            messages.append(("system", system))
        messages.append(("human", self.convert_varibales(self.prompt, inputs)))
        return messages
    








    

    


    def run_schema(self, inputs:Iinputs):
        llm = self.get_llm(inputs)
        _schema =  self.create_schema()
        PROMPT = self.create_messages(inputs)
        answer = llm.with_schema(PROMPT, _schema, self.name)
        return answer


    def create_schema(self) -> Dict:
        if not self.schema_data:
            raise ValueError("schema_dict is empty")

        schema = {
            "name":self.name,
            "description":"use this schema to create structured output and return object",
            "parameters":{
                "type":"object",    
                "properties":{},
                "required":[]
            },
        }

        for s in self.schema_data:
            obj_prop = {
                "type":s.type,
                "description":s.description,
            }
            if s.is_enum:
                obj_prop["enum"] = s.enums

            schema["parameters"]["properties"][s.key] = obj_prop

            if s.required:
                schema["parameters"]["required"].append(s.key)
        return schema
        
        



import json
from typing import List
from OfficelyTestTeam2.run import run

from team_builder.nodes.enums import MessageType
from team_builder.nodes.interface import IHistory




def get_team(path:str)->dict:
    with open(path, "r") as f:
        return json.load(f)


if __name__ == "__main__":
    team = get_team("team.json")
    verbose = True
    chat_history:List[IHistory] = []

    while True:
        query = input("User: ")
        res = run(team, verbose, chat_history, query)
        chat_history.append(IHistory(text=query, type=MessageType.INBOUND))
        chat_history.append(IHistory(text=res, type=MessageType.OUTBOUND))
        print(f"AI:{res}")


from .factory import factory_llm, is_openAI, is_anthropic
from .bedrock import BedrockLLM
from .bedrock_tools import BedrockTools
from .openai import OpenAILLM
from .enums import LLMType
from .tokenizer import Tokenizer


__all__ = ['factory_llm',
           'BedrockLLM', 
           'BedrockTools', 
           'OpenAILLM', 
           'LLMType', 
           'Tokenizer', 
            'is_openAI',
            'is_anthropic'
        ]



from typing import Dict, Optional, List
from abc import ABC, abstractmethod
import re
from pydantic.v1 import BaseModel


from team_builder.nodes.interface import Iinputs

from .enums import NodeType


class Node(BaseModel, ABC):
    id:str
    name:str
    description:Optional[str] = None
    _answer:Optional[str] = None


    @property
    @abstractmethod
    def type(self) -> NodeType:
        """return the type of the node"""
        
    @abstractmethod
    def execute(self, inputs:Iinputs) -> str | Dict :
        """execute the node""" 

    @classmethod
    @abstractmethod
    def from_kwargs(cls):
        """create the node from kwargs"""



    def validate_placeholders(self, template:str, data:dict):
        def get_nested_data(data, keys):
            for key in keys:
                try:
                    data = data[key]
                except (TypeError, KeyError):
                    return None
            return data
        
        #placeholders = re.findall(r'\{(.*?)\}', template)
        placeholders = re.findall(r'\{\{(.*?)\}\}', template)
        missing = []

        # Process each placeholder
        for placeholder in placeholders:
            keys = placeholder.split('.')
            result = get_nested_data(data, keys)
            # Check if the result is None or if it is still a placeholder
            if result is None:
                missing.append(placeholder)

        if missing:
            raise ValueError(f"Missing placeholders: {', '.join(missing)} in Node: {self.name}")

    def convert_varibales(self, text:str, inputs:Iinputs) -> str:
        self.validate_placeholders(text, inputs.as_dict())
        _inputs = inputs.as_dict_str()

        def find_value(keys:List, value:Dict):
            try:
                for key in keys:
                    if "[" in key:
                        key, index = key[:-1].split("[")
                        value = value[key][int(index)]
                    else:
                        value = value[key]
                return value
            except (KeyError, IndexError, ValueError):
                return None
            
        def replace_placeholders_with_symbol(text: str, data: Dict, symbol: str):
            placeholders = re.findall(rf"\{symbol}(\w+)", text)
            for placeholder in placeholders:
                if placeholder in data:
                    text = text.replace(
                        f"{symbol}{placeholder}", str(data[placeholder])
                    )
            return text
        
        text = replace_placeholders_with_symbol(text, _inputs, "$")
        while matches := list(re.finditer(r"\{\{([\w\.\[\]]+)\}\}", text)):
            for match in reversed(matches):
                key_str = match.group(1)
                keys = key_str.split(".")
                value = find_value(keys, _inputs)
                if value is not None:
                    text = text[:match.start()] + str(value) + text[match.end():]
                else:
                    text = text[:match.start()] + text[match.end():]
        return text






from typing import Dict, List, Optional

from OfficelyTestTeam2.generator import ThreadedGenerator
from .enums import OperatorValue
from team_builder.nodes.enums import NodeType
from team_builder.nodes.filter.interfaces import ICondition, IFilterObject
from team_builder.nodes.interface import Iinputs
from team_builder.nodes.node import Node

from team_builder.nodes.filter.enums import DataType




class FilterNode(Node):
    """this node use filter the childs """
    childs: List["Node"] = []
    filter_object: IFilterObject




    @property
    def type(self):
        return NodeType.FILTER

    @classmethod
    def from_kwargs(cls, **kwargs):
        return cls(
            id = kwargs["id"],
            name = "Filter",
            filter_object = kwargs['filterObject']
        )

    def execute(self, inputs:Iinputs, g:Optional[ThreadedGenerator] = None):
        next_child = None
        for group in self.filter_object.groups:
            _bool = False
            temp_bool = False
            conditions = group.conditions
            for i, con in enumerate(conditions):
                field, value = self.convert_varibales(con.field, inputs), self.convert_varibales(con.value, inputs)
                if not field and not value:
                    field = None
                try:
                    field, value = self.convert_data_type(str(field), value, con)
                except ValueError:
                    continue
                temp_bool = self.logical_by_operator(field, value, con.operator.label)
                if i != 0:
                    if conditions[i - 1].logicalOperator == "AND":
                        _bool = _bool and temp_bool
                    else:
                        _bool = _bool or temp_bool
                else:
                    _bool = temp_bool
            if _bool:
                next_child = self.__next_child(group.id)
                self.send_verbose(self.name, next_child.name, g)
                break
        else:
            next_child = self.__next_child(self.filter_object.elseID)
            self.send_verbose(self.name, next_child.name, g)
        return next_child


    def __next_child(self, _id:str):
        return next((item for item in self.childs if item.id == _id ))
    
    def send_verbose(self, parent:str, child:str, g:ThreadedGenerator=None):
        if g:
            g.send(f"<code style='color:red'>{parent} -> {child}</code>\n\n", False)
        
    

    def convert_data_type(self, field:str, value:str, condition:ICondition):
        try:
            data_type = condition.operator.dataType
            match data_type:
                case DataType.TEXT:
                    new_field, new_value = str(field).strip(), str(value).strip()
                case DataType.NUMBER:
                    new_field, new_value = float(field), float(value)
                case DataType.BASIC:
                    if field == condition.field:
                        new_field = None
                case _:
                    raise ValueError(f"DataType {data_type} not supported")
        except ValueError as e:
            raise e
        else:
            return new_field, new_value
        
    def logical_by_operator(self, field, value, operator:OperatorValue):
        match operator:
            case OperatorValue.EQUAL.value:  # done
                return field == value
            case OperatorValue.EQUAL_INSENSETIVE.value:  # done
                return str(field).lower() == str(value).lower()
            case OperatorValue.NOT_EQUAL.value:  # done
                return field != value
            case OperatorValue.NOT_EQUAL_INSENSETIVE.value:  # done
                return str(field).lower() != str(value).lower()
            case OperatorValue.GREATER_THAN.value:
                return field > value
            case OperatorValue.LESS_THAN.value:
                return field < value
            case OperatorValue.GREATER_THAN_OR_EQUAL_TO.value:
                return field >= value
            case OperatorValue.LESS_THAN_OR_EQUAL_TO.value:
                return field <= value
            case OperatorValue.STARTS_WITH.value:
                return str(field).startswith(value)
            case OperatorValue.STARTS_WITH_INSENSETIVE.value:
                return str(field).lower().startswith(str(value).lower())
            case OperatorValue.NOT_STARTS_WITH.value:
                return not str(field).startswith(value)
            case OperatorValue.NOT_STARTS_WITH_INSENSETIVE.value:
                return not str(field).lower().startswith(str(value).lower())
            case OperatorValue.ENDS_WITH.value:
                return str(field).endswith(value)
            case OperatorValue.ENDS_WITH_INSENSETIVE.value:
                return str(field).lower().endswith(str(value).lower())
            case OperatorValue.NOT_ENDS_WITH.value:
                return not str(field).endswith(value)
            case OperatorValue.NOT_ENDS_WITH_INSENSETIVE.value:
                return not str(field).lower().endswith(str(value).lower())
            case OperatorValue.CONTAINS.value:
                return value in field
            case OperatorValue.CONTAINS_INSENSETIVE.value:
                return str(value).lower() in str(field).lower()
            case OperatorValue.NOT_CONTAINS.value:
                return value not in field
            case OperatorValue.NOT_CONTAINS_INSENSETIVE.value:
                return str(value).lower() not in str(field).lower()
            case OperatorValue.EXISTS.value:
                return field is not None
            case OperatorValue.NOT_EXISTS.value:
                return field is None
        return False
        


        
        


from typing import Dict, List, Optional, Sequence, cast, Any
from pydantic.v1 import BaseModel, Field, root_validator
from colorama import Fore, Style
from team_builder.nodes.enums import NodeType
from OfficelyTestTeam2.generator import ThreadedGenerator
from team_builder.nodes.filter.filter_node import FilterNode
from team_builder.nodes.interface import Iinputs, InputItem
from team_builder.nodes.jump import JumpNode
from team_builder.nodes.llm_node import LLMNode
from .nodes.node import Node
from .llms import Tokenizer

class Edge(BaseModel):
    source: str
    target: str

class Graph(BaseModel):
    nodes: Sequence[Node]
    edges: List[Edge]
    tokenizer:Optional[Tokenizer]=None
    g:Optional[ThreadedGenerator] = None
    verbose:bool = True
    inputs: Iinputs = Field(default_factory=lambda: Iinputs())
    current_node:Optional[str] =  ""


    @root_validator(pre=True)
    def set_inputs_verbose(cls, values):
        verbose = values.get('verbose', True)
        values['inputs'] = Iinputs(verbose=verbose)
        return values




    def __send_verbose(self, key:str, answer:Any, _type:NodeType):
        if _type in {NodeType.HISTORY, NodeType.QUESTION}:
            return

        color_mapping = {
            NodeType.LLM: "#5B61EB",
            NodeType.PYTHON: "green",
            NodeType.RETRIVER: "#B9B9B9",
            NodeType.JUMP: "purple",
        }
        default_color = "black"

        color = color_mapping.get(_type, default_color)
        if self.g:
            self.g.send(f"<code style='color:{color}'>{key}: {answer}</code>\n\n", False) 

    def __add_input(self, key:str, answer:Any, _type:NodeType, print_verbose:bool=True):
        new_input = InputItem(key=key, value=answer, type=_type)
        self.inputs[key] = new_input # pylint: disable=unsupported-assignment-operation
        if self.verbose and print_verbose:
            print(Fore.MAGENTA + ("----------" * 3))
            print(Fore.GREEN + f"{key} - {answer}" + Style.RESET_ALL)
            # if not last_node:
            #     self.__send_verbose(key, answer, _type)

    def run(self, query:str, chat_history:List[Dict]):
        self.__add_input("question", query, NodeType.QUESTION, False)
        self.__add_input("chat_history", chat_history, NodeType.HISTORY, False)
        node = self.__run_node()
        return self.inputs.get_value(node.name)

    def __run_node(self):
        try:
            next_node = self.__next_node()
            while next_node is not None:
                node = next_node
                self.current_node = node.id  # Correctly use the property
                next_node = self.__next_node()
                match node.type:
                    case NodeType.LLM:
                        llm_node = cast(LLMNode, node)
                        llm_node.g = self.g
                        llm_node.tokenizer = self.tokenizer
                        if next_node is None  and llm_node.g:
                            llm_node.g.is_answer = True
                    case NodeType.FILTER:
                        filter_node = cast(FilterNode, node)
                        filter_node.childs = [item for item in self.nodes if item.id in self.get_edge_child(node)]
                        next_node = filter_node.execute(self.inputs, self.g if self.verbose else None)
                        continue
                    case NodeType.JUMP:
                        jump_node = cast(JumpNode, node)
                        next_node = self.find_node(jump_node.jump_to_id)
                        if next_node is None:
                            raise ValueError(f"Jump node '{jump_node.name}' not found")
                        self.__add_input(jump_node.name, f" => {next_node.name}", jump_node.type, print_verbose=(next_node is not None))
                        continue
                res_node = node.execute(self.inputs)
                self.__add_input(node.name, res_node, node.type, print_verbose=(next_node is not None))
            return node
        except Exception as e:
            if node:    
                err = f"Error in **{node.name}**: {e}"
            else: 
                err = str(e) 
            raise Exception(err) 
        
    def find_node(self, node_id:str):
        return next((item for item in self.nodes if item.id == node_id), None)

    def __next_node(self, node_id:Optional[str]=None) -> Node | None:
        if node_id:
            return self.find_node(node_id)
        if not self.current_node:
            return self.get_start_node()
        try:
            edge:Edge = next((item for item in self.edges if item.source == self.current_node))
        except StopIteration:
            return None
        return next((item for item in self.nodes if item.id == edge.target))

    def get_edge_child(self, node:Node):
        return [x.target for x in self.edges if x.source == node.id]
    
    def get_start_node(self):
        for node in self.nodes:
            if not any(edge.target == node.id for edge in self.edges):
                return node
        raise ValueError("Graph has no start node")





import json
from operator import itemgetter
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
    Callable
)

from langchain_core.callbacks import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models import LanguageModelInput

from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    BaseMessageChunk,
    SystemMessage,
)

from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
from langchain_core.pydantic_v1 import BaseModel, Field
from langchain_core.tools import BaseTool

from langchain_community.chat_models.bedrock import BedrockChat, ChatPromptAdapter
from typing import Tuple
import defusedxml.ElementTree as DET

from langchain_anthropic.chat_models import convert_to_anthropic_tool, _format_messages

from langchain_core.runnables import (
    Runnable,
    RunnableMap,
    RunnablePassthrough,
)

from typing import Any, List, Optional, Type, TypedDict, cast

from typing import Any, List, Optional, Type, TypedDict, cast

from langchain_core.messages import BaseMessage
from langchain_core.output_parsers import BaseGenerationOutputParser
from langchain_core.outputs import ChatGeneration, Generation
from langchain_core.pydantic_v1 import BaseModel


class _ToolCall(TypedDict):
    name: str
    args: dict
    id: str
    index: int


class ToolsOutputParser(BaseGenerationOutputParser):
    first_tool_only: bool = False
    args_only: bool = False
    pydantic_schemas: Optional[List[Type[BaseModel]]] = None

    class Config:
        extra = "forbid"

    def parse_result(self, result: List[Generation], *, partial: bool = False) -> Any:
        """Parse a list of candidate model Generations into a specific format.

        Args:
            result: A list of Generations to be parsed. The Generations are assumed
                to be different candidate outputs for a single model input.

        Returns:
            Structured output.
        """
        if not result or not isinstance(result[0], ChatGeneration):
            return None if self.first_tool_only else []
        tool_calls: List = _extract_tool_calls(result[0].message)
        if self.pydantic_schemas:
            tool_calls = [self._pydantic_parse(tc) for tc in tool_calls]
        elif self.args_only:
            tool_calls = [tc["args"] for tc in tool_calls]
        else:
            pass

        if self.first_tool_only:
            return tool_calls[0] if tool_calls else None
        else:
            return tool_calls

    def _pydantic_parse(self, tool_call: _ToolCall) -> BaseModel:
        cls_ = {schema.__name__: schema for schema in self.pydantic_schemas or []}[
            tool_call["name"]
        ]
        return cls_(**tool_call["args"])


def _extract_tool_calls(msg: BaseMessage) -> List[_ToolCall]:
    if isinstance(msg.content, str):
        return []
    tool_calls = []
    for i, block in enumerate(cast(List[dict], msg.content)):
        if block["type"] != "tool_use":
            continue
        tool_calls.append(
            _ToolCall(name=block["name"], args=block["input"], id=block["id"], index=i)
        )
    return tool_calls



SYSTEM_PROMPT_FORMAT = """In this environment you have access to a set of tools you can use to answer the user's question.

You may call them like this:
<function_calls>
<invoke>
<tool_name>$TOOL_NAME</tool_name>
<parameters>
<$PARAMETER_NAME>$PARAMETER_VALUE</$PARAMETER_NAME>
...
</parameters>
</invoke>
</function_calls>

Here are the tools available:
<tools>
{formatted_tools}
</tools>"""  # noqa: E501

TOOL_FORMAT = """<tool_description>
<tool_name>{tool_name}</tool_name>
<description>{tool_description}</description>
<parameters>
{formatted_parameters}
</parameters>
</tool_description>"""

TOOL_PARAMETER_FORMAT = """<parameter>
<name>{parameter_name}</name>
<type>{parameter_type}</type>
<description>{parameter_description}</description>
</parameter>"""


def _get_type(parameter: Dict[str, Any]) -> str:
    if "type" in parameter:
        return parameter["type"]
    if "anyOf" in parameter:
        return json.dumps({"anyOf": parameter["anyOf"]})
    if "allOf" in parameter:
        return json.dumps({"allOf": parameter["allOf"]})
    return json.dumps(parameter)


def get_system_message(tools: List[Dict]) -> str:
    tools_data: List[Dict] = [
        {
            "tool_name": tool["name"],
            "tool_description": tool["description"],
            "formatted_parameters": "\n".join(
                [
                    TOOL_PARAMETER_FORMAT.format(
                        parameter_name=name,
                        parameter_type=_get_type(parameter),
                        parameter_description=parameter.get("description"),
                    )
                    for name, parameter in tool["parameters"]["properties"].items()
                ]
            ),
        }
        for tool in tools
    ]
    tools_formatted = "\n".join(
        [
            TOOL_FORMAT.format(
                tool_name=tool["tool_name"],
                tool_description=tool["tool_description"],
                formatted_parameters=tool["formatted_parameters"],
            )
            for tool in tools_data
        ]
    )
    return SYSTEM_PROMPT_FORMAT.format(formatted_tools=tools_formatted)


def _xml_to_dict(t: Any) -> Union[str, Dict[str, Any]]:
    # Base case: If the element has no children, return its text or an empty string.
    if len(t) == 0:
        return t.text or ""

    # Recursive case: The element has children. Convert them into a dictionary.
    d: Dict[str, Any] = {}
    for child in t:
        if child.tag not in d:
            d[child.tag] = _xml_to_dict(child)
        else:
            # Handle multiple children with the same tag
            if not isinstance(d[child.tag], list):
                d[child.tag] = [d[child.tag]]  # Convert existing entry into a list
            d[child.tag].append(_xml_to_dict(child))
    return d


def _xml_to_function_call(invoke: Any, tools: List[Dict]) -> Dict[str, Any]:
    name = invoke.find("tool_name").text
    arguments = _xml_to_dict(invoke.find("parameters"))

    # make list elements in arguments actually lists
    filtered_tools = [tool for tool in tools if tool["name"] == name]
    if len(filtered_tools) > 0 and not isinstance(arguments, str):
        tool = filtered_tools[0]
        for key, value in arguments.items():
            if key in tool["parameters"]["properties"]:
                if "type" in tool["parameters"]["properties"][key]:
                    if tool["parameters"]["properties"][key][
                        "type"
                    ] == "array" and not isinstance(value, list):
                        arguments[key] = [value]
                    if (
                        tool["parameters"]["properties"][key]["type"] != "object"
                        and isinstance(value, dict)
                        and len(value.keys()) == 1
                    ):
                        arguments[key] = list(value.values())[0]

    return {
        "function": {
            "name": name,
            "arguments": json.dumps(arguments),
        },
        "type": "function",
    }


def _xml_to_tool_calls(elem: Any, tools: List[Dict]) -> List[Dict[str, Any]]:
    """
    Convert an XML element and its children into a dictionary of dictionaries.
    """
    invokes = elem.findall("invoke")

    return [_xml_to_function_call(invoke, tools) for invoke in invokes]


class BedrockTools(BedrockChat):

    _xmllib: Any = Field(default=DET)


    def bind_tools(
        self,
        tools: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool]],
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        """Bind tool-like objects to this chat model.

        Args:
            tools: A list of tool definitions to bind to this chat model.
                Can be  a dictionary, pydantic model, callable, or BaseTool. Pydantic
                models, callables, and BaseTools will be automatically converted to
                their schema dictionary representation.
            **kwargs: Any additional parameters to bind.
        """
        formatted_tools = [convert_to_anthropic_tool(tool) for tool in tools]
        return self.bind(tools=formatted_tools, **kwargs)

    def with_structured_output(
        self,
        schema: Union[Dict, Type[BaseModel]],
        *,
        include_raw: bool = False,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, Union[Dict, BaseModel]]:
        llm = self.bind_tools([schema])
        if isinstance(schema, type) and issubclass(schema, BaseModel):
            output_parser = ToolsOutputParser(
                first_tool_only=True, pydantic_schemas=[schema]
            )
        else:
            output_parser = ToolsOutputParser(first_tool_only=True, args_only=True)

        if include_raw:
            parser_assign = RunnablePassthrough.assign(
                parsed=itemgetter("raw") | output_parser, parsing_error=lambda _: None
            )
            parser_none = RunnablePassthrough.assign(parsed=lambda _: None)
            parser_with_fallback = parser_assign.with_fallbacks(
                [parser_none], exception_key="parsing_error"
            )
            return RunnableMap(raw=llm) | parser_with_fallback
        else:
            return llm | output_parser

    def _format_params(
        self,
        *,
        messages: List[BaseMessage],
        **kwargs: Any,
    ) -> Tuple[Optional[str], Union[List[Dict], str]]:
        tools: List[Dict] = kwargs.get("tools", None)
        # experimental tools are sent in as part of system prompt, so if
        # both are set, turn system prompt into tools + system prompt (tools first)
        if tools:
            tool_system = get_system_message(tools)

            if messages[0].type == "system":
                sys_content = messages[0].content
                new_sys_content = f"{tool_system}\n\n{sys_content}"
                messages = [SystemMessage(content=new_sys_content), *messages[1:]]
            else:
                messages = [SystemMessage(content=tool_system), *messages]

        provider = self._get_provider()
        if provider == 'anthropic':
            return ChatPromptAdapter.format_messages(provider, messages)
        return None, ChatPromptAdapter.convert_messages_to_prompt(provider, messages)

        #return super()._format_params(messages=messages, stop=stop, **kwargs)

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        # streaming not supported for functions
        result = self._generate(
            messages=messages, stop=stop, run_manager=run_manager, **kwargs
        )
        to_yield = result.generations[0]
        chunk = ChatGenerationChunk(
            message=cast(BaseMessageChunk, to_yield.message),
            generation_info=to_yield.generation_info,
        )
        if run_manager:
            run_manager.on_llm_new_token(
                cast(str, to_yield.message.content), chunk=chunk
            )
        yield chunk

    async def _astream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        # streaming not supported for functions
        result = await self._agenerate(
            messages=messages, stop=stop, run_manager=run_manager, **kwargs
        )
        to_yield = result.generations[0]
        chunk = ChatGenerationChunk(
            message=cast(BaseMessageChunk, to_yield.message),
            generation_info=to_yield.generation_info,
        )
        if run_manager:
            await run_manager.on_llm_new_token(
                cast(str, to_yield.message.content), chunk=chunk
            )
        yield chunk

    # def _format_output(self, text: Any, **kwargs: Any) -> ChatResult:
    #     """Format the output of the model, parsing xml as a tool call."""
    #     #text = data.content[0].text
    #     tools = kwargs.get("tools", None)

    #     additional_kwargs: Dict[str, Any] = {}

    #     if tools:
    #         # parse out the xml from the text
    #         try:
    #             # get everything between <function_calls> and </function_calls>
    #             start = text.find("<function_calls>")
    #             end = text.find("</function_calls>") + len("</function_calls>")
    #             xml_text = text[start:end]

    #             xml = DET.fromstring(xml_text)
    #             additional_kwargs["tool_calls"] = _xml_to_tool_calls(xml, tools)
    #             text = ""
    #         except Exception:
    #             pass

    #     return ChatResult(
    #         generations=[
    #             ChatGeneration(
    #                 message=AIMessage(content=text, additional_kwargs=additional_kwargs)
    #             )
    #         ],
    #         #llm_output=data,
    #     )

    def _format_output(self, text: Any, additional_kwargs: Any) -> ChatResult:
        #data_dict = data.model_dump()
        # content = data_dict["content"]
        # llm_output = {
        #     k: v for k, v in data_dict.items() if k not in ("content", "role", "type")
        # }
        # if len(content) == 1 and content[0]["type"] == "text":
        #     msg = AIMessage(content=content[0]["text"])
        # else:
        return ChatResult(
            generations=[
                ChatGeneration(
                    message=AIMessage(content=text, additional_kwargs=additional_kwargs)
                )
            ],
            #llm_output=data,
        )
    
    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        completion = ""

        if self.streaming:
            for chunk in self._stream(messages, stop, run_manager, **kwargs):
                completion += chunk.text
        else:
            system = None
            formatted_messages:Optional[List[Dict]] = None
            prompt:Optional[str]= None
            params: Dict[str, Any] = {**kwargs}
            system, formatted_messages = _format_messages(messages)
            # if not system:
            #     prompt = str(messages_prompt)
            # else:
            #     assert isinstance(messages_prompt, list), "messages_prompt must be a list of BaseMessage"
            #     formatted_messages = messages_prompt
            
            if stop:
                params["stop_sequences"] = stop

            completion = self._prepare_input_and_invoke(
                prompt=None,
                stop=stop,
                run_manager=run_manager,
                system=system,
                messages=formatted_messages,
                **params,
            )

        
        return self._format_output(completion[0], params)


def _tools_in_params(params: dict) -> bool:
    return "tools" in params or (
        "extra_body" in params and params["extra_body"].get("tools")
    )




import threading
from typing import Dict, List, Optional, Tuple

from OfficelyTestTeam2.generator import ThreadedGenerator
from team_builder.graph import Edge, Graph
from team_builder.llms.tokenizer import Tokenizer, get_team_tokenizer_callback
import re

from team_builder.nodes.factory import factory_node
from team_builder.nodes.interface import IHistory, InputItem
import json

def run(
        team: Dict,
        verbose:bool,
        chat_history: List[IHistory],
        query: str,
        g: Optional[ThreadedGenerator] = None,
    ) -> str:
        try:
            team_res = None

            tokenizer = Tokenizer()
            with get_team_tokenizer_callback(tokenizer) as cb:
                team_res = run_team(chat_history, query, team, g, verbose, tokenizer)
                if verbose and g and cb.data:
                    g.send(f"\n\n<pre>{cb}</pre>")
        except Exception as e:  # This will catch other types of exceptions
            err_msg = str(e) if verbose else "there is error in agent, please try again."
            if bool(g):
                g.send(err_msg)
            else:
                return err_msg
            if not verbose:
               raise e

        finally:
            if bool(g):
                full_answer = re.sub(r'<(code|pre).*?>.*?</\1>', '', g.full_answer, flags=re.DOTALL).strip()
                if not full_answer and g.team and team_res:
                    g.send(str(team_res))
                g.close()
            
        return str(team_res)

            
def run_team(
            chat_history, 
            query, 
            team:Dict,
            g:Optional[ThreadedGenerator]=None, 
            verbose:bool=False,
            tokenizer:Optional[Tokenizer]=None
        )->str: 

        if g:
            g.team = True
        if not chat_history:
            return team["settings"]["startMessage"]

        nodes = [factory_node(node['data']['type'], **node['data'], ) for node in team['nodes'] if node['type'] == 'workflow']
        edges = [Edge(source=edge['source'], target=edge['target']) for edge in team['edges'] if edge['type'] == 'workflow']

        graph = Graph(nodes=nodes, edges=edges, g=g, verbose=verbose, tokenizer=tokenizer)
        res = graph.run(query, chat_history)
        return res


def arun(team:Dict, verbose:bool, chat_history:List, query: str):
    try:
        g = ThreadedGenerator()
        threading.Thread(target=run, args=(team, verbose, chat_history, query)).start()
        return g
    except Exception as e:
        raise e


def stream_response(team:Dict, verbose:bool, chat_history:List, query: str):
        try:
            ai_response = arun(team, verbose, chat_history, query)
            def generate():
                try:
                    for item in ai_response:  # Iterate over the ThreadedGenerator instance
                        yield json.dumps(item, ensure_ascii=False) + "\n"

                except Exception as e:
                    print(f"Stream Response: {e}")
                    replay = "there is error in agent, please try again."
                    yield json.dumps(replay, ensure_ascii=False) + "\n"
            return generate()

        except Exception as e:
            raise e


from sys import api_version
from typing import Dict, List, Tuple, Optional

from pydantic.v1 import root_validator, validator
from OfficelyTestTeam2.stream_handler import StreamHandler
from .llm import LLM
from langchain_openai import AzureChatOpenAI
import os
from OfficelyTestTeam2.CustomException import RootValidatorException



class AzureLLM(LLM):
    azure_deployment_name: Optional[str]
    streaming:bool = True


    @root_validator(pre=True)
    def check_azure_deployment_name(cls, values):
        if not values.get('azure_deployment_name'):
            raise RootValidatorException("Missing Azure Deployment Name")
        return values



    @property
    def base(self):
        callbacks = []
        if self.tokenizer:
            self.tokenizer.model = self.model
            callbacks.append(self.tokenizer)
        is_answer = False
        if self.g:
            callbacks.append(StreamHandler(gen=self.g, with_final=False))
            is_answer = self.g.is_answer
            
        api_version = "2024-05-01-preview"

        return AzureChatOpenAI(
            azure_deployment=self.azure_deployment_name,
            api_version=api_version,
            temperature=self.temperature,
            streaming=is_answer and self.streaming, 
            callbacks=callbacks
        )
            
    def with_schema(self, prompt:List[Tuple[str, str]], schema:Dict, name:str):
        self.streaming = False
        return self.base.with_structured_output(schema).with_config({"run_name":name}).invoke(prompt)
    




from typing import Any, Dict, List, Optional


from team_builder.nodes.interface import Iinputs

from .enums import NodeType
from .node import Node

from langchain_community.vectorstores import Pinecone
from langchain.retrievers import ContextualCompressionRetriever
from langchain.schema import format_document
from langchain.prompts.prompt import PromptTemplate
from langchain_core.documents import Document
from langchain_cohere import CohereRerank
import os





class RetriverNode(Node):
    question:str
    index:str
    namespace:str
    embedding: Any
    k:Optional[int]
    n:Optional[int]
    rerank_score: Optional[float]
    filter: List[Dict]
    loaders: List[Dict]



    @property
    def type(self):
        return NodeType.RETRIVER
    
    @classmethod
    def from_kwargs(cls, **kwargs):
        return cls(
            id = kwargs["id"],
            name = kwargs["name"],
            description = kwargs.get("description"),
            question = kwargs["question"],
            index = "basic",
            namespace = os.environ['NAMESPACE'],
            k = kwargs.get("k", 10),
            n = kwargs.get("n", 4),
            rerank_score = kwargs.get("rerank_score"),
            embedding = "embedding",
            filter = kwargs.get("filter", []),
            loaders = kwargs.get("loaders", [])

        )


    def execute(self, inputs:Iinputs) -> str | Dict | List :
        docs = self._rernk(self.get_relevant_documents(self.convert_varibales(self.question, inputs)))
        metadata = []
        for x in docs:
            metadata.append({"content":x.page_content, "metadata": x.metadata})
        format_docs = self.format_docs(docs)
        string_docs = self.get_doc_string(format_docs)
        return {"context": string_docs, "documents": metadata}
        #return self.get_doc_string(docs)

    # @property
    # def Embeddings(self):
    #     match self.embeding_model:
    #         case EmbeddingModel.TITAN_1:
    #             if not self.aws_config:
    #                 aws_config = {"region_name":'us-east-1'}
    #             bedrock = boto3.client(service_name='bedrock-runtime', **aws_config)
    #             return BedrockEmbeddings(client=bedrock, model_id=self.embeding_model)
    #         case EmbeddingModel.OPEN_AI_SMALL:
    #             return OpenAIEmbeddings(model=self.embeding_model)
    #         case _:
    #             if self.azure_config:
    #                 return AzureOpenAIEmbeddings(**self.azure_config)
    #             return OpenAIEmbeddings()


    
    def get_relevant_documents(self, query:str) -> List[Document]:
        _list = []
        loader_urls = []
        for loader in self.loaders:
            if loader['type'] == 'URL':
                for url in loader['urls']:
                    loader_urls.append(url['url'])
            else:
                loader_urls.append(loader['id'])

        # for x in self.filter:
        #     if x['id'] in loader_urls:
        #         if x['type'] == LoaderType.URL:
        #             _list.append(x['id'])
        #         else:
        #             _list.append(f"{self.namespace}/{x['id']}")

        filter = {'source':{ "$in": _list }} if _list else {}
        vectorstore = Pinecone.from_existing_index(self.index, self.embedding, namespace=self.namespace)
        retriever = vectorstore.as_retriever(search_kwargs={"k":self.k, "filter":filter})
        compressor = CohereRerank(model="rerank-multilingual-v2.0", top_n=self.n)
        return ContextualCompressionRetriever(base_compressor=compressor, base_retriever=retriever).invoke(query)
    

    
    def _rernk(self, docs:List[Document]) -> List[Document]:
        if not self.rerank_score:
            return docs
        return [doc for doc in docs if doc.metadata["relevance_score"] >= self.rerank_score]
    
    def format_docs(self, docs:List[Document]) -> List[Dict]:
        document_prompt = PromptTemplate.from_template(template="{page_content}")
        return [format_document(doc, document_prompt) for doc in docs] #type: ignore
    
    def get_doc_string(self, docs, document_separator="\n\n"):
        return document_separator.join(docs)

    
        


        
            















        






from typing import List, Optional
from pydantic.v1 import BaseModel

from .enums import OperatorValue
from team_builder.nodes.filter.enums import DataType






class IOperatorInfo(BaseModel):
    label: OperatorValue
    dataType: DataType

class ICondition(BaseModel):
    field: str
    operator: IOperatorInfo
    value: str
    logicalOperator:Optional[str] = None



class IGroupCondtions(BaseModel):
    id: str
    conditions: List[ICondition]

class IFilterObject(BaseModel):
    groups: List[IGroupCondtions]
    elseID:str


from setuptools import setup, find_packages

setup(
    name='OfficelyTestTeam2',
    version='0.1.3',
    author='Your Name',
    author_email='your.email@example.com',
    description='A brief description of your package',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),  # Automatically includes all packages
    include_package_data=True,  # Include additional files as specified in MANIFEST.in
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)



from enum import Enum

class NodeType(str, Enum):
    LLM = 'llm'
    RETRIVER = 'retriver'
    FILTER = 'filter'
    PYTHON = 'python',
    TEXT = 'text' 
    JUMP = 'jump'
    ########## for DictInputsType
    QUESTION = "question"
    HISTORY = "history"



class SchemaType(str, Enum):
    STRING = "string"
    BOOL = "boolean"
    NUMBER = "number"
    FLOAT = "float"

# class DictInputsType(str, Enum):
#     TEXT = "text"
#     HISTORY = "history"

    
# class MessageType(str, Enum):
#     AI_OUTBOUND = "ai-outbound"
#     OUTBOUND = "outbound"
#     HUMAN = "human"
#     AI = "ai"

class EmbeddingModel(str, Enum):
    AZURE = "AZURE",
    ADA2 = "text-embedding-ada-002",
    OPEN_AI_SMALL = "text-embedding-3-small"
    TITAN_1 = "amazon.titan-embed-text-v1",
    TITAN_2 = "amazon.titan-embed-text-v2:0",
    COHERE_MULTY = "cohere.embed-multilingual-v3"

class MessageType(str, Enum):
    OUTBOUND = "outbound"
    INBOUND = "inbound"

